import json
import logging
from typing import Type

from langchain_core.prompts import ChatPromptTemplate
from pydantic import BaseModel

from duck_agents.llm.session_factory import SessionFactory

_logger = logging.getLogger(__name__)

_system_prompt_template = (
    "{task_description}\n"
    "The user will provide the input for your task in structured JSON format."  
    "The input adheres to the following JSON schema:\n"
    "{input_schema}"
)

class BaseAgent:
    def __init__(
            self,
            llm_session_factory: SessionFactory,
            task_description: str,
            input_schema: Type[BaseModel],
            output_schema: Type[BaseModel]):

        self.llm_session_factory = llm_session_factory
        self.input_schema = input_schema
        self.task_description = task_description
        input_schema_json = json.dumps(self.input_schema.model_json_schema())
        self.prompt_template = ChatPromptTemplate.from_messages(
            [
                ("system", _system_prompt_template),
                ("user", "{data}")
            ]
        )
        self.llm_session = self.llm_session_factory.create_session(output_schema=output_schema)
        self.chain = self.prompt_template | self.llm_session
        self.agent_type = type(self).__name__


    def prompt(self, data: BaseModel) -> BaseModel:
        return self.chain.invoke(
            {
                "data": data.model_dump_json(),
                "task_description": self.task_description,
                "input_schema": self.input_schema.model_json_schema()
            }
        )


from typing import Type

import jsonref
from langchain_google_genai import ChatGoogleGenerativeAI
from pydantic import BaseModel, SecretStr

from duck_agents.llm.session_factory import SessionFactory

# Assumes that either env variable GOOGLE_API_KEY is set or the api key is passed as arg
class GeminiSessionFactory(SessionFactory):
    def __init__(self, model_name, api_key: SecretStr = None):
        self.model_name = model_name
        self.api_key = api_key

    def create_session(self, output_schema: Type[BaseModel] = None, temperature: float = 0):
         kwargs = {
             "model": self.model_name,
             "temperature": temperature,
         }
         if self.api_key is not None: kwargs["google_api_key"] = self.api_key
         # gemini pydantic nesting issue https://medium.com/@andreasantoro.pvt/make-gemini-json-output-stricter-4feccf570d8c
         # and solution from https://github.com/pydantic/pydantic/issues/889 to unnest the model
         output_json_schema = jsonref.replace_refs(output_schema.model_json_schema())
         return ChatGoogleGenerativeAI(**kwargs).with_structured_output(schema=output_json_schema)



from typing import Type

from langchain_ollama import OllamaLLM
from pydantic import BaseModel

from duck_agents.llm.session_factory import SessionFactory

class OLlamaSessionFactory(SessionFactory):
    def __init__(self, url: str, model_name: str):
        self.url = url
        self.model_name = model_name

    def create_session(self, output_schema: Type[BaseModel] = None, temperature: float = 0):
        # TODO: implement workaround, since OLLAMA doesn't support function calling (or langchain didn't implement it yet)
        return (OllamaLLM(model=self.model_name, temperature=temperature, base_url=self.url)
                  .with_structured_output(schema=output_schema))


import logging
from typing import Type

from pydantic import BaseModel

_logger = logging.getLogger(__name__)

class SessionFactory:
    def create_session(self, output_schema: Type[BaseModel] = None, temperature: float = 0):
        _logger.warning(
            "Trying to create a session with SessionFactory instance." 
            "Subclass this class and provide your own implementation of create_session.")